from collections import OrderedDict
from functools import partial
import json
import logging
from pathlib import Path

from addict import Dict as Addict
from atomicwrites import atomic_write
import dateutil.parser

from . import core, jncweb, spec, utils
from .trio_utils import bag

logger = logging.getLogger(__package__)
console = utils.getConsole()


# TODO change => in App folder for windows
# TODO config file there too (login, passsword) as requested by someone no GH tracker
DEFAULT_CONFIG_FILEPATH = Path.home() / ".jncep" / "tracked.json"


class TrackConfigManager:
    def __init__(self, config_file_path=None):
        if not config_file_path:
            self.config_file_path = DEFAULT_CONFIG_FILEPATH
        else:
            if config_file_path is Path:
                self.config_file_path = config_file_path
            else:
                self.config_file_path = Path(config_file_path)
            # TODO check read write permission / is file etc...

    # TODO async
    def read_tracked_series(self):
        try:
            with self.config_file_path.open() as json_file:
                # Explicit ordereddict (although should be fine without
                # since Python >= 3.6 dicts are ordered ; spec since 3.7)
                data = json.load(json_file, object_pairs_hook=OrderedDict)
                data = Addict(data)
                return self._convert_to_latest_format(data)
        except FileNotFoundError:
            # first run ?
            return Addict({})

    def write_tracked_series(self, tracked):
        self._ensure_config_dirpath_exists()
        with atomic_write(str(self.config_file_path.resolve()), overwrite=True) as f:
            f.write(json.dumps(tracked, sort_keys=True, indent=2))

    def _convert_to_latest_format(self, data):
        converted = {}
        # while at it convert from old format
        # legacy format for tracked parts : just the part instead of object
        # with keys part, name
        # key is slug
        # TODO rename "name" field into "title"
        for series_url_or_slug, value in data.items():
            if not isinstance(value, dict):
                series_slug = series_url_or_slug
                series_url = jncweb.url_from_series_slug(series_slug)
                # low effort way to get some title
                name = series_slug.replace("-", " ").title()
                value = Addict({"name": name, "part": value})
                converted[series_url] = value
            else:
                converted[series_url_or_slug] = value

        converted_b = {}
        for legacy_series_url, value in converted.items():
            new_series_url = jncweb.to_new_website_series_url(legacy_series_url)
            converted_b[new_series_url] = value

        return converted_b

    def _ensure_config_dirpath_exists(self):
        self.config_file_path.parent.mkdir(parents=False, exist_ok=True)


async def fill_meta_last_part(session, series):
    await core.fill_volumes_meta(session, series)
    volumes = series.volumes

    if volumes:
        # just the last
        last_volume = volumes[-1]
        parts = await core.fetch_parts_meta_for_volume(session, last_volume.volume_id)
        last_volume.parts = parts
        for part in parts:
            part.volume = last_volume
            part.series = last_volume.series

    return series


async def track_series(session, tracked_series, series):
    await fill_meta_last_part(session, series)

    last_part = None
    if series.volumes:
        last_volume = series.volumes[-1]
        if last_volume.parts:
            last_part = last_volume.parts[-1]

    # record current last part + name
    if not last_part:
        # no parts yet
        pn = 0
        # 0000-... not a valid date so 1111-...
        pdate = "1111-11-11T11:11:11.111Z"

        console.info(
            f"The series '[highlight]{series.raw_data.title}[/]' is now tracked, "
            "starting [highlight]from the beginning[/]",
            style="success",
        )
    else:
        pn = spec.to_relative_spec_from_part(last_part)
        pdate = last_part.raw_data.launch

        relative_part = spec.to_relative_spec_from_part(last_part)
        part_date = dateutil.parser.parse(last_part.raw_data.launch)
        part_date_formatted = part_date.strftime("%b %d, %Y")
        console.info(
            f"The series '[highlight]{series.raw_data.title}[/]' is now tracked, "
            f"starting after part [highlight]{relative_part} [{part_date_formatted}]"
            f"[/]",
            style="success",
        )

    series_url = jncweb.url_from_series_slug(series.raw_data.slug)
    tracked_series[series_url] = Addict(
        {
            "part_date": pdate,
            "part": pn,  # now just for show
            "name": series.raw_data.title,
        }
    )


async def sync_series_forward(session, follows, tracked_series, is_delete):
    # sync local tracked series based on remote follows
    new_synced = []
    del_synced = []

    async def do_track(jnc_resource):
        series = await core.resolve_series(session, jnc_resource)
        await track_series(session, tracked_series, series)

        series_url = jncweb.url_from_series_slug(series.raw_data.slug)
        new_synced.append(series_url)

    tasks = []
    for jnc_resource in follows:
        if jnc_resource.url in tracked_series:
            continue
        tasks.append(partial(do_track, jnc_resource))

    # result doesn't matter ; just for the exceptions
    await bag(tasks)

    if is_delete:
        followed_index = {f.url: f for f in follows}
        # list() to avoid dictionary changed size during iteration
        for series_url, series_data in list(tracked_series.items()):
            if series_url not in followed_index:
                del tracked_series[series_url]

                console.warning(
                    f"The series '[highlight]{series_data.name}[/]' is no longer "
                    "tracked"
                )

                del_synced.append(series_url)

    return new_synced, del_synced


async def sync_series_backward(session, follows, tracked_series, is_delete):
    # sync remote follows based on locally tracked series
    new_synced = []
    del_synced = []

    async def do_follow(jnc_resource):
        console.info(f"Fetch metadata for '{jnc_resource}'...")
        series = await core.resolve_series(session, jnc_resource)
        series_id = series.series_id
        title = series.raw_data.title

        console.info(f"Follow '{title}'...")
        await session.api.follow_series(series_id)

        new_synced.append(series_url)

    followed_index = {f.url: f for f in follows}
    tasks = []
    for series_url in tracked_series:
        # series_url is the latest URL format (same as the follows)
        if series_url in followed_index:
            continue
        jnc_resource = jncweb.resource_from_url(series_url)
        tasks.append(partial(do_follow, jnc_resource))

    if is_delete:

        async def do_undollow(jnc_resource):
            # use the follow_raw_data: to avoid another call to the API
            series_id = jnc_resource.follow_raw_data.id
            title = jnc_resource.follow_raw_data.title
            console.warning(f"Unfollow '{title}'...")
            await session.api.unfollow_series(series_id)

            del_synced.append(jnc_resource.url)

        for jnc_resource in follows:
            if jnc_resource.url not in tracked_series:
                tasks.append(partial(do_undollow, jnc_resource))

    # no result needed ; just for the exceptions
    await bag(tasks)

    return new_synced, del_synced



import logging

import click

from .cli.epub import generate_epub
from .cli.track import track_series
from .cli.update import update_tracked
from .utils import module_info, setup_logging

logger = logging.getLogger(module_info())


@click.group(
    help="Simple command-line tool to generate EPUB files for J-Novel Club pre-pub "
    "novels"
)
@click.option(
    "-d",
    "--debug",
    "is_debug",
    is_flag=True,
    help=("Flag to activate debug mode"),
    required=False,
)
def main(is_debug):
    setup_logging(is_debug)


main.add_command(generate_epub)
main.add_command(track_series)
main.add_command(update_tracked)

if __name__ == "__main__":
    main()



import logging
import re

import attr

logger = logging.getLogger(__name__)

RANGE_SEP = ":"

SERIES = "ALL_SERIES"
VOLUME = "ALL_VOLUME"
PART = "SINGLE_PART"
START_OF_VOLUME = "START_OF_VOLUME"
END_OF_VOLUME = "END_OF_VOLUME"
START_OF_SERIES = "START_OF_SERIES"
END_OF_SERIES = "END_OF_SERIES"


@attr.s
class Single:
    type_ = attr.ib()
    spec = attr.ib()

    def has_volume(self, volume) -> bool:
        if self.type_ == SERIES:
            return True
        elif self.type_ == VOLUME:
            return self.spec == volume.num
        # part
        vn, _ = self.spec
        return volume.num == vn

    def has_part(self, ref_part) -> bool:
        # assume has_volume is True if has_part is checked
        if self.type_ in (SERIES, VOLUME):
            return True
        # part
        _, pn = self.spec
        return ref_part.num_in_volume == pn


@attr.s
class Interval:
    start = attr.ib()
    end = attr.ib()

    def has_volume(self, volume) -> bool:
        if self.start == START_OF_SERIES:
            # spec is a part (START_OF / END_OF cannot happen together)
            vn, _ = self.end.spec
            return volume.num <= vn

        vn, _ = self.start.spec
        if self.end == END_OF_SERIES:
            return volume.num >= vn

        vn2, _ = self.end.spec
        return vn <= volume.num <= vn2

    def has_part(self, ref_part) -> bool:
        if self.start == START_OF_SERIES:
            # spec is a part
            vn, pn = self.end.spec
            if ref_part.volume.num < vn:
                return True
            # same volume
            return ref_part.num_in_volume <= pn

        vn, pn = self.start.spec
        if self.end == END_OF_SERIES:
            if ref_part.volume.num > vn:
                return True
            # same volume
            return ref_part.num_in_volume >= pn

        vn2, pn2 = self.end.spec
        if vn < ref_part.volume.volume_num < vn2:
            return True
        if ref_part.volume.num == vn and ref_part.num_in_volume >= pn:
            return True
        if ref_part.volume.num == vn2 and ref_part.num_in_volume <= pn2:
            return True

        return False


def to_relative_spec_from_part(part):
    volume_number = part.volume.num
    part_number = part.num_in_volume
    return f"{volume_number}.{part_number}"


def analyze_part_specs(part_specs):
    """v(.p):v2(.p) or v(.p): or :v(.p) or v(.p) or :"""

    part_specs = part_specs.strip()

    if part_specs == RANGE_SEP:
        return Single(SERIES, None)

    return _analyze_volume_part_specs(part_specs)


def _analyze_volume_part_specs(part_specs):
    sides = part_specs.split(RANGE_SEP)
    if len(sides) > 2:
        raise ValueError("Multiple ':' in part specs")

    reg = r"^\s*(\d+)(?:\.(\d+))?\s*$"
    if len(sides) == 1:
        # not a range: single part
        m = re.match(reg, sides[0])
        if not m:
            raise ValueError(
                "Specification must be a of the form 'vol[.part]' (part is optional)"
            )
        fv = int(m.group(1))
        if m.group(2):
            # only the part specified
            fp = int(m.group(2))
            return Single(PART, (fv, fp))
        else:
            # full volume
            return Single(VOLUME, fv)

    # range
    m1 = re.match(reg, sides[0])
    m2 = re.match(reg, sides[1])
    if (
        (not m1 and not m2)
        # left side not valid
        or (not m1 and len(sides[0]) > 0)
        # right side not valid
        or (not m2 and len(sides[1]) > 0)
    ):
        msg = (
            "Part specification must be vol[.part]:vol[.part] or vol[.part]: or "
            ":vol[.part]"
        )
        raise ValueError(msg)

    if m1:
        fv = int(m1.group(1))
        if m1.group(2):
            fp = int(m1.group(2))
        else:
            fp = START_OF_VOLUME
        start = Single(PART, (fv, fp))
    else:
        start = START_OF_SERIES

    if m2:
        lv = int(m2.group(1))
        if m2.group(2):
            lp = int(m2.group(2))
        else:
            lp = END_OF_VOLUME
        end = Single(PART, (lv, lp))
    else:
        end = END_OF_SERIES

    # both sides
    return Interval(start, end)



from collections import namedtuple
from datetime import datetime, timezone
from functools import partial
from html.parser import HTMLParser
import json
import logging
import os
import re
import sys
import time
from typing import List

from addict import Dict as Addict
import attr
import dateutil.parser
import trio

from . import epub, jnclabs, jncweb, spec, utils
from .model import Image, Part, Series, Volume
from .trio_utils import bag
from .utils import deep_freeze, to_safe_filename

logger = logging.getLogger(__name__)
console = utils.getConsole()


EpubGenerationOptions = namedtuple(
    "EpubGenerationOptions",
    [
        "output_dirpath",
        "is_by_volume",
        "is_extract_images",
        "is_extract_content",
        "is_not_replace_chars",
        "style_css_path",
    ],
)


@attr.s
class IdentifierSpec:
    type_ = attr.ib()
    volume_id = attr.ib(None)
    part_id = attr.ib(None)

    def has_volume(self, volume) -> bool:
        if self.type_ == spec.SERIES:
            return True

        return self.volume_id == volume.volume_id

    def has_part(self, part) -> bool:
        # assumes has_volume already checked with the volume_id
        if self.type_ in (spec.SERIES, spec.VOLUME):
            return True

        return self.part_id == part.part_id


class JNCEPSession:
    def __init__(self, email, password):
        self.api = jnclabs.JNCLabsAPI()
        self.email = email
        self.password = password
        self.now = datetime.now(tz=timezone.utc)

    async def __aenter__(self) -> "JNCEPSession":
        await self.login(self.email, self.password)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        console.stop_status()

        await self.logout()
        return False

    async def login(self, email, password):
        console.status(f"Login with email '[highlight]{email}[/]'...")
        token = await self.api.login(email, password)

        emoji = ""
        if console.is_advanced():
            emoji = "\u26A1 "
        console.info(f"{emoji}Logged in with email '[highlight]{email}[/]'")
        console.status("...")
        return token

    async def logout(self):
        if self.api.is_logged_in:
            try:
                console.info("Logout...")
                await self.api.logout()
            except (trio.MultiError, Exception) as ex:
                logger.debug(f"Error logout: {ex}", exc_info=sys.exc_info())


async def create_epub(series, volumes, parts, epub_generation_options):
    book_details = process_series(series, volumes, parts, epub_generation_options)

    if epub_generation_options.is_extract_content:
        await extract_content(parts, epub_generation_options)

    if epub_generation_options.is_extract_images:
        await extract_images(parts, epub_generation_options)

    for book_details_i in book_details:
        output_filename = to_safe_filename(book_details_i.title) + ".epub"
        output_filepath = os.path.join(
            epub_generation_options.output_dirpath, output_filename
        )
        # TODO write to memory then async fs write here ? (uses epublib
        # which is sync anyway)
        # or trio.to_thread.run_sync or inside
        epub.output_epub(
            output_filepath, book_details_i, epub_generation_options.style_css_path
        )

        # laughing face
        emoji = ""
        if console.is_advanced():
            emoji = "\U0001F600 "
        console.info(
            f"{emoji}Success! EPUB generated in '{output_filepath}'!",
            style="success",
        )


def process_series(
    series, volumes, parts, options: EpubGenerationOptions
) -> epub.BookDetails:

    # prepare content
    for part in parts:
        content_for_part = part.content
        if not options.is_not_replace_chars:
            content_for_part = _replace_chars(content_for_part)

        # some parts do not have an image
        if part.images:
            imgs = part.images
            content_for_part = _replace_image_urls(content_for_part, imgs)

        part.epub_content = content_for_part

    if options.is_by_volume:
        book_details = []
        for volume in volumes:
            volume_parts = [part for part in parts if part.volume is volume]
            volume_details = _process_single_epub_content(
                series, [volume], volume_parts
            )
            book_details.append(volume_details)
    else:
        book_details = [_process_single_epub_content(series, volumes, parts)]

    return book_details


def _process_single_epub_content(series, volumes, parts):
    # order of volumes and parts must match

    # representative volume
    repr_volume = volumes[0]
    author = _extract_author(repr_volume.raw_data.creators)

    # representative part
    repr_part = parts[0]
    volume_num = repr_part.volume.num
    part_num = repr_part.num_in_volume

    # TODO in case multiple volumes, do not set volume num? or do
    # something else
    collection = epub.CollectionMetadata(
        series.raw_data.legacyId, series.raw_data.title, volume_num
    )

    # cover can be None (handled in epub gen proper)
    cover_image = repr_volume.cover

    contents = [part.epub_content for part in parts]

    if len(parts) == 1:
        # single part
        part = parts[0]
        volume_num = part.volume.num
        part_num = part.num_in_volume

        suffix = ""
        if _is_part_final(part):
            suffix = " [Final]"

        title = f"{part.raw_data.title}{suffix}"
        # single part => single volume: part numbers relative to
        # that volume
        toc = [f"Part {part_num}"]
    else:
        volume_index = set([v.num for v in volumes])
        if len(volume_index) > 1:
            volume_nums = sorted(list(volume_index))
            volume_nums = [str(vn) for vn in volume_nums]
            volume_nums = ", ".join(volume_nums[:-1]) + " & " + volume_nums[-1]
            title_base = f"{series.raw_data.title}: Volumes {volume_nums}"

            volume_num0 = parts[0].volume.num
            part_num0 = parts[0].num_in_volume
            volume_num1 = parts[-1].volume.num
            part_num1 = parts[-1].num_in_volume

            # check only last part in the epub
            suffix = ""
            if _is_part_final(parts[-1]):
                suffix = " - Final"

            part_nums = (
                f"Parts {volume_num0}.{part_num0} to "
                f"{volume_num1}.{part_num1}{suffix}"
            )

            toc = [part.raw_data.title for part in parts]
            title = f"{title_base} [{part_nums}]"
        else:
            volume = volumes[0]
            title_base = volume.raw_data.title

            toc = [f"Part {part.num_in_volume}" for part in parts]

            part_num0 = parts[0].num_in_volume
            part_num1 = parts[-1].num_in_volume

            is_complete = _is_volume_complete(volume, parts)
            if is_complete:
                part_nums = "Complete"
            else:
                # check the last part in the epub
                suffix = ""
                if _is_part_final(parts[-1]):
                    suffix = " - Final"
                part_nums = f"Parts {part_num0} to {part_num1}{suffix}"

            title = f"{title_base} [{part_nums}]"

    identifier = series.raw_data.slug + str(int(time.time()))

    images = [img for part in parts for img in part.images]

    book_details = epub.BookDetails(
        identifier, title, author, collection, cover_image, toc, contents, images
    )

    return book_details


def _is_part_final(part):
    volume = part.volume
    if volume.parts_count is None:
        # assume not final
        return False
    return part.num_in_volume == volume.parts_count


def _is_volume_complete(volume, parts):
    # need parts as args : the requested parts that will be included in the final
    # epub
    if volume.parts_count is None:
        # assume not complete
        return False
    return volume.parts_count == len(parts)


async def extract_images(parts, epub_generation_options):
    async with trio.open_nursery() as n:
        for part in parts:
            images = part.images

            for image in images:
                # change filename to something more readable since visible to
                # user
                _, ext = os.path.splitext(image.local_filename)
                img_filename = (
                    to_safe_filename(part.raw_data.title)
                    # extension at the end
                    + f"_Image_{image.order_in_part}{ext}"
                )
                img_filepath = os.path.join(
                    epub_generation_options.output_dirpath, img_filename
                )
                n.start_soon(_write_bytes, img_filepath, image.content)


async def extract_content(parts, epub_generation_options):
    async with trio.open_nursery() as n:
        for part in parts:
            content = part.content
            content_filename = to_safe_filename(part.raw_data.title) + ".html"
            content_filepath = os.path.join(
                epub_generation_options.output_dirpath, content_filename
            )
            n.start_soon(_write_str, content_filepath, content)


def _extract_author(creators, default="Unknown Author"):
    for creator in creators:
        if creator.role == "AUTHOR":
            return creator.name
    return default


def _replace_chars(content):
    # both the chars to replace and replacement are hardcoded
    # U+2671 => East Syriac Cross
    # U+25C6 => Black Diamond
    # U+1F3F6 => Black Rosette
    # U+25C7 => White Diamond
    # U+2605 => Black star
    chars_to_replace = ["\u2671", "\u25C6", "\U0001F3F6", "\u25C7", "\u2605"]
    replacement_char = "**"
    regex = "|".join(chars_to_replace)
    content = re.sub(regex, replacement_char, content)
    return content


def _replace_image_urls(content, images: List[Image]):
    for image in images:
        # the filename relative to the epub content root
        # the file will be added to the Epub archive
        content = content.replace(image.url, image.local_filename)

    return content


def all_parts_meta(series):
    return [part for volume in series.volumes if volume.parts for part in volume.parts]


async def to_part_spec(session, jnc_resource):
    if jnc_resource.resource_type == jncweb.RESOURCE_TYPE_SERIES:
        return IdentifierSpec(spec.SERIES)

    elif jnc_resource.resource_type == jncweb.RESOURCE_TYPE_VOLUME:
        if jnc_resource.is_new_website:
            # for volume on new website => where is a tuple (series_slug,
            # volume num)
            series_slug, volume_number = jnc_resource.slug

            volumes = await fetch_volumes_for_series(session.api, series_slug)

            volume_index = volume_number - 1
            if volume_index not in range(len(volumes)):
                raise jncweb.BadWebURLError(
                    f"Incorrect volume number in URL: {jnc_resource.url}"
                )

            volume = volumes[volume_index]
        else:
            volume_slug = jnc_resource.slug
            volume = await session.api.fetch_data("volumes", volume_slug)

        return IdentifierSpec(spec.VOLUME, volume.legacyId)

    elif jnc_resource.resource_type == jncweb.RESOURCE_TYPE_PART:
        tasks = [
            partial(session.api.fetch_data, "parts", jnc_resource.slug),
            partial(session.api.fetch_data, "parts", jnc_resource.slug, "volume"),
        ]
        part, volume = await bag(tasks)
        return IdentifierSpec(spec.PART, volume.legacyId, part.legacyId)


async def resolve_series(session, jnc_resource):
    if jnc_resource.resource_type == jncweb.RESOURCE_TYPE_SERIES:
        series_slug = jnc_resource.slug
        series_raw_data = await session.api.fetch_data("series", series_slug)
    elif jnc_resource.resource_type == jncweb.RESOURCE_TYPE_VOLUME:
        if jnc_resource.is_new_website:
            series_slug, _ = jnc_resource.slug
            series_raw_data = await session.api.fetch_data("series", series_slug)
        else:
            volume_slug = jnc_resource.slug
            series_raw_data = await session.api.fetch_data(
                "volumes", volume_slug, "serie"
            )
    elif jnc_resource.resource_type == jncweb.RESOURCE_TYPE_PART:
        part_slug = jnc_resource.slug
        series_raw_data = await session.api.fetch_data("parts", part_slug, "serie")

    series = Series(series_raw_data, series_raw_data.legacyId)
    return series


async def fill_meta(session, series, volume_callback=None):
    await fill_volumes_meta(session, series)
    volumes = series.volumes

    await fill_parts_meta_for_volumes(session, volumes, volume_callback)

    # similar to what we got from the original JNC API
    return series


async def fetch_volumes_meta(session, series_id):
    volumes_raw_data = await fetch_volumes_for_series(session, series_id)

    volumes = []
    for i, volume_raw_data in enumerate(volumes_raw_data):
        volume_id = volume_raw_data.legacyId
        volume_num = i + 1

        volume = Volume(volume_raw_data, volume_id, volume_num)
        volumes.append(volume)

    return volumes


async def fill_volumes_meta(session, series):
    volumes = await fetch_volumes_meta(session, series.series_id)
    series.volumes = volumes
    for volume in volumes:
        volume.series = series


async def fill_parts_meta_for_volumes(session, volumes, volume_callback=None):
    volumes_meta = []
    tasks = []
    for volume in volumes:
        if volume_callback and not volume_callback(volume):
            continue
        tasks.append(partial(fetch_parts_meta_for_volume, session, volume.volume_id))
        volumes_meta.append(volume)
    all_parts = await bag(tasks)

    for i, parts in enumerate(all_parts):
        volume = volumes_meta[i]

        volume.parts = parts
        for part in parts:
            part.volume = volume
            part.series = volume.series


async def fetch_parts_meta_for_volume(session, volume_id):
    parts_raw_data = await fetch_parts_for_volume(session, volume_id)
    parts = []

    for i, part_raw_data in enumerate(parts_raw_data):
        part_id = part_raw_data.legacyId
        part_num = i + 1
        part = Part(part_raw_data, part_id, part_num)
        parts.append(part)

    return parts


async def fetch_volumes_for_series(session, series_id):
    volumes = [
        volume
        async for volume in session.api.paginate(
            partial(session.api.fetch_data, "series", series_id, "volumes")
        )
    ]
    return volumes


async def fetch_parts_for_volume(session, volume_id):
    parts = [
        part
        async for part in session.api.paginate(
            partial(session.api.fetch_data, "volumes", volume_id, "parts")
        )
    ]
    return parts


async def fetch_content(session, parts):
    tasks = []
    for part in parts:
        tasks.append(partial(fetch_content_and_images_for_part, session, part.part_id))
    contents = await bag(tasks)

    parts_content = {}
    for i, content_image in enumerate(contents):
        part = parts[i]
        parts_content[part.part_id] = content_image
    return parts_content


def is_part_available(now, part):
    if part.raw_data.preview:
        return True

    if part.series.raw_data.catchup:
        return True

    expiration_data = dateutil.parser.parse(part.raw_data.expiration)
    return expiration_data > now


async def fetch_content_and_images_for_part(session, part_id):
    # TODO catch error + event => in case expires between checking before and
    # running this (case to check)
    content = await session.api.fetch_content(part_id, "data.xhtml")
    img_urls = extract_image_urls(content)
    if len(img_urls) > 0:
        tasks = [partial(fetch_image, session, img_url) for img_url in img_urls]
        images = await bag(tasks)

        # filter images with download error
        images = list(filter(None, images))
        for i, image in enumerate(images):
            image.order_in_part = i + 1

    else:
        images = []

    return content, images


async def fetch_image(session, img_url):
    try:
        img_bytes = await session.api.fetch_url(img_url)
        image = Image(img_url, img_bytes)
        image.local_filename = _local_image_filename(image)
        return image
    except (trio.MultiError, Exception) as ex:
        console.error(f"Error downloading image with URL: '{img_url}'")
        logger.debug(f"Error downloading image: {ex}", exc_info=sys.exc_info())
        # TODO still create the Image object with empty content ?
        # the original link to the image will stay in the EPUB
        # some EPUB reader may be able to donwload them
        return None


def _local_image_filename(image):
    # unique name to use for the image inside the EPUB
    # ext is almost always .jpg but sometimes it is .jpeg
    # splitext  works fine with a url
    root, ext = os.path.splitext(image.url)
    new_local_filename = to_safe_filename(root) + ext
    return new_local_filename


class ImgUrlParser(HTMLParser):
    def __init__(self):
        super().__init__()

        self.img_urls = []

    def handle_starttag(self, tag, tag_attrs):
        if tag == "img":
            for tag_attr in tag_attrs:
                if tag_attr[0] == "src":
                    self.img_urls.append(tag_attr[1])
                    break


def extract_image_urls(content):
    parser = ImgUrlParser()
    parser.feed(content)
    return parser.img_urls


async def fetch_covers(session, volumes):
    tasks = []
    for volume in volumes:
        # fetch the cover url low res image as indicated in the metadata
        # used as a fallback in case the high res is not found
        tasks.append(partial(fetch_lowres_cover_for_volume, session, volume))
        tasks.append(partial(fetch_cover_image_from_parts, session, volume.parts))

    covers = await bag(tasks)
    # even index are lowres
    lowres_covers = covers[::2]
    # odd
    hires_covers = covers[1::2]

    candidate_covers = zip(hires_covers, lowres_covers)
    volumes_cover = {}
    for i, candidate_cover in enumerate(candidate_covers):
        volume = volumes[i]
        hires, lowres = candidate_cover
        # priority to hires
        # note : lowres can also be none if failure => handled in epub gen
        volumes_cover[volume.volume_id] = hires if hires else lowres

    return volumes_cover


async def fetch_lowres_cover_for_volume(session, volume):
    try:
        cover_url = volume.raw_data.cover.coverUrl
        cover = await fetch_image(session, cover_url)
        return cover
    except (trio.MultiError, Exception) as ex:
        logger.debug(
            f"Error fetch_lowres_cover_for_volume: {ex}", exc_info=sys.exc_info()
        )
        return None


async def fetch_cover_image_from_parts(session, parts):
    # need content so only available parts
    parts = [part for part in parts if is_part_available(session.now, part)]
    if not parts:
        return None

    try:
        # the cover image is almost always on the 1st part
        # very rarely on the second
        # so check those 2 first in parallel (need the content); then if
        # doesn't succeed, check the rest
        first_2_parts = parts[:2]
        rest_parts = parts[2:]

        async def fetch_highres_image_maybe(session, part_id):
            content = await session.api.fetch_content(part_id, "data.xhtml")
            return _candidate_cover_image(content)

        for batch_parts in [first_2_parts, rest_parts]:
            tasks = []
            for part in batch_parts:
                tasks.append(partial(fetch_highres_image_maybe, session, part.part_id))
            candidate_urls = await bag(tasks)
            cover = await _fetch_one_candidate_image(session, candidate_urls)
            if cover:
                return cover
        return None

    except (trio.MultiError, Exception) as ex:
        logger.debug(
            f"Error fetching hi res cover images: {ex}", exc_info=sys.exc_info()
        )
        # lowres cover will be used (unless it fails too)
        return None


async def _fetch_one_candidate_image(session, candidate_urls):
    for candidate_url in candidate_urls:
        if candidate_url:

            if "cover" not in candidate_url and "cvr" not in candidate_url:
                # TODO check the cover format on old series
                logger.debug(
                    "The hires cover candidate url doesn't look like a cover URL"
                )

            cover = await fetch_image(session, candidate_url)
            # TODO check dimension ? but all the images in the interior have
            # hi res
            # TODO check the presence of colors => images in the interior except the
            # cover are B&W
            return cover
    return None


class StopHTMLParsing(Exception):
    pass


class CoverImgUrlParser(HTMLParser):
    def __init__(self):
        super().__init__()

        self.candidate_cover_url = None

        # this is to handle covers : special significance for
        # first image before any text
        self.is_after_body = False
        self.is_after_first_text_in_body = False

    def handle_starttag(self, tag, tag_attrs):
        if tag == "img":
            self._handle_img(tag_attrs)
            raise StopHTMLParsing()

        if tag == "body":
            self.is_after_body = True

    def _handle_img(self, tag_attrs):
        for tag_attr in tag_attrs:
            if tag_attr[0] == "src":
                self.candidate_cover_url = tag_attr[1]
                break

    def handle_data(self, text):
        if self.is_after_body:
            # to avoid new line and spaces
            text = text.strip()
            if text:
                raise StopHTMLParsing()


def _candidate_cover_image(content):
    try:
        parser = CoverImgUrlParser()
        parser.feed(content)
    except StopHTMLParsing:
        pass
    url = parser.candidate_cover_url
    if not url:
        return None
    return url


def relevant_volumes_for_cover(volumes, is_by_volume):
    if not is_by_volume:
        volumes_cover = [volumes[0]]
    else:
        volumes_cover = volumes
    return volumes_cover


def relevant_volumes_and_parts_for_content(series, part_filter):
    # some volumes may be empty after checking the parts => so getting
    # the volumes from the parts
    volumes_to_download = {}
    parts_to_download = []
    for volume in series.volumes:
        if not volume.parts:
            continue
        for part in volume.parts:
            if part_filter(part):
                parts_to_download.append(part)
                volumes_to_download[part.volume.volume_id] = volume

    # dict is ordered from py 3.6 so no need to sort
    volumes_to_download = list(volumes_to_download.values())

    return volumes_to_download, parts_to_download


async def fill_covers_and_content(session, cover_volumes, content_parts):
    tasks = [
        partial(fetch_content, session, content_parts),
        partial(fetch_covers, session, cover_volumes),
    ]
    contents, covers = await bag(tasks)

    for part in content_parts:
        if part.part_id in contents:
            content, images = contents[part.part_id]
            part.content = content
            part.images = images

    for volume in cover_volumes:
        if volume.volume_id in covers:
            volume.cover = covers[volume.volume_id]


async def fetch_jncweb_page_react_data(session, series_slug):
    page_html = await session.api.fetch_jnc_webpage(series_slug)
    start_tag = '<script id="__NEXT_DATA__" type="application/json">'
    end_tag = "</script>"
    start_index = page_html.find(start_tag)
    if start_index == -1:
        return None
    start_index += len(start_tag)
    end_index = page_html.find(end_tag, start_index)
    react_props = page_html[start_index:end_index]
    data = Addict(json.loads(react_props))
    deep_freeze(data)
    return data


async def fill_num_parts_for_volumes(session, series, volumes):
    props = await fetch_jncweb_page_react_data(session, series.raw_data.slug)

    def find_volumes(obj):
        if isinstance(obj, dict):
            if "volumes" in obj:
                return obj.volumes
            for item in obj.values():
                volumes_attr = find_volumes(item)
                if volumes_attr:
                    return volumes_attr
        elif isinstance(obj, list):
            for item in obj:
                volumes_attr = find_volumes(item)
                if volumes_attr:
                    return volumes_attr
        return None

    volumes_data = find_volumes(props)
    if not volumes_data:
        return

    volumes_parts_count_index = {}
    for volume_data in volumes_data:
        # partCount seems to be there even for older series
        if "partCount" in volume_data:
            volumes_parts_count_index[volume_data.id] = volume_data.partCount

    for volume in volumes:
        if volume.volume_id in volumes_parts_count_index:
            volume.parts_count = volumes_parts_count_index[volume.volume_id]


async def _write_bytes(filepath, content):
    async with await trio.open_file(filepath, "wb") as img_f:
        await img_f.write(content)


async def _write_str(filepath, content):
    async with await trio.open_file(filepath, "w", encoding="utf-8") as img_f:
        await img_f.write(content)



import logging

import attr
from ebooklib import epub
import importlib_resources as imres

from .model import Image

logger = logging.getLogger(__name__)


@attr.s
class BookDetails:
    identifier = attr.ib()
    title = attr.ib()
    author = attr.ib()
    collection = attr.ib()
    cover_image = attr.ib()
    toc = attr.ib()
    contents = attr.ib()
    images = attr.ib()


@attr.s
class CollectionMetadata:
    collection_id = attr.ib()
    collection_title = attr.ib()
    position = attr.ib()


DEFAULT_STYLE_CSS_PATH = "res/style.css"
DEFAULT_STYLE_CSS = None


def read_default_style_css():
    global DEFAULT_STYLE_CSS
    DEFAULT_STYLE_CSS = (
        imres.files(__package__).joinpath(DEFAULT_STYLE_CSS_PATH).read_text()
    )


def get_css(style_css_path):
    if style_css_path:
        with open(style_css_path, "r", encoding="utf-8", errors="ignore") as f:
            return f.read()

    if not DEFAULT_STYLE_CSS:
        read_default_style_css()

    return DEFAULT_STYLE_CSS


def output_epub(output_filepath, book_details: BookDetails, style_css_path=None):
    lang = "en"
    book = epub.EpubBook()
    book.set_identifier(book_details.identifier)
    book.set_title(book_details.title)
    book.set_language(lang)
    book.add_author(book_details.author)

    # metadata for series GH issue #9
    collection_meta = book_details.collection
    book.add_metadata(
        "OPF",
        "belongs-to-collection",
        collection_meta.collection_title,
        {"property": "belongs-to-collection", "id": collection_meta.collection_id},
    )
    book.add_metadata(
        "OPF",
        "collection-type",
        "series",
        {"property": "collection-type", "refines": f"#{collection_meta.collection_id}"},
    )

    # as position, set the volume number of the first part in the epub
    # in Calibre, display 1 (I) if not set so a bit better
    book.add_metadata(
        "OPF",
        "group-position",
        str(collection_meta.position),
        {"property": "group-position", "refines": f"#{collection_meta.collection_id}"},
    )

    if book_details.cover_image:
        content = book_details.cover_image.content
        # in case cover image also present in content, use the file name
        # (same URL => same local filename)
        cover_image_filename = book_details.cover_image.local_filename
    else:
        # the lib handles that semi-gracefully (doesn't crash)
        # may look broken in epub reader
        # TODO handle problem with missing cover => use dummy jpeg
        content = None
        # dummy file name
        cover_image_filename = "cover.jpg"

    # TODO why not True ? check
    book.set_cover(cover_image_filename, content, False)

    style = get_css(style_css_path)

    css = epub.EpubItem(
        uid="style", file_name="book.css", media_type="text/css", content=style
    )
    book.add_item(css)

    # TODO cf why not True ? above
    cover_page = epub.EpubHtml(title="Cover", file_name="cover.xhtml", lang=lang)
    cover_page.content = f'<img src="{cover_image_filename}" alt="cover" />'
    cover_page.add_item(css)
    book.add_item(cover_page)

    image: Image
    for image in book_details.images:
        # do not add if already added through cover or problems when writing:
        # "Duplicate name" warning from epublib + maybe issue in the epub zip structure
        if image.local_filename == cover_image_filename:
            continue
        img = epub.EpubImage()
        img.file_name = image.local_filename
        # TODO always ? check ?
        img.media_type = "image/jpeg"
        img.content = image.content
        book.add_item(img)

    chapters = []
    for i, content in enumerate(book_details.contents):
        c = epub.EpubHtml(
            title=book_details.toc[i], file_name=f"chap_{i}.xhtml", lang=lang
        )
        # explicit encoding to bytes or some issue with lxml on some platforms (PyDroid)
        # some message about USC4 little endian not supported
        c.content = content.encode("utf-8")
        c.add_item(css)
        book.add_item(c)
        chapters.append(c)

    book.toc = chapters

    book.add_item(epub.EpubNcx())
    book.add_item(epub.EpubNav())

    book.spine = [cover_page, "nav", *chapters]

    epub.write_epub(output_filepath, book, {})



from functools import partial
import logging
import sys

import attr
import dateutil
import trio

from . import core, jncweb, model, spec, track, utils
from .trio_utils import bag

logger = logging.getLogger(__package__)
console = utils.getConsole()


@attr.s
class UpdateResult:
    series = attr.ib(None)
    is_error = attr.ib(False)
    is_updated = attr.ib(None)
    is_considered = attr.ib(True)
    # to indicate a series with expired parts only
    # will set to latest part or will always have error
    # if stalled
    is_force_set_updated = attr.ib(False)


async def update_url_series(
    session,
    jnc_url,
    epub_generation_options,
    tracked_series,
    is_sync,
    new_synced,
    is_whole_volume,
):
    # for single url => if error no catch : let it crash and report to the user
    jnc_resource = jncweb.resource_from_url(jnc_url)
    series_meta = await core.resolve_series(session, jnc_resource)

    series_url = jncweb.url_from_series_slug(series_meta.raw_data.slug)

    if series_url not in tracked_series:
        console.warning(
            f"The series '[highlight]{series_meta.raw_data.title}[/]' is not tracked! "
            f"Use the 'jncep track add' command first."
        )
        return

    if is_sync:
        # not very useful but make it possible
        # only consider newly synced series if --sync used
        # to mirror case with no URL argument
        if series_url not in new_synced:
            console.warning(
                f"The series '[highlight]{series_meta.raw_data.title}[/]' is not "
                "among the tracked series added from syncing. Use 'jncep update' "
                "without --sync."
            )
            return

    series_details = tracked_series[series_url]

    # the series has just been synced so force EPUB gen from start
    is_force_from_beginning = is_sync
    update_result = await _create_epub_for_new_parts(
        session,
        series_details,
        series_meta,
        epub_generation_options,
        is_whole_volume,
        is_force_from_beginning,
    )

    if update_result.is_updated:
        emoji = ""
        if console.is_advanced():
            emoji = "\u2714 "
        console.info(
            f"{emoji}The series '[highlight]{series_meta.raw_data.title}[/]' has "
            "been updated!",
            style="success",
        )
    else:
        console.info(
            f"The series '[highlight]{series_meta.raw_data.title}[/]' is already up "
            "to date!",
            style="success",
        )

    is_tracking_updated = _update_tracking_data(
        series_details, series_meta, update_result
    )

    return is_tracking_updated


async def update_all_series(
    session,
    epub_generation_options,
    tracked_series,
    is_sync,
    new_synced,
    is_whole_volume,
):
    series_details_a = []
    tasks = []
    for series_url, series_details in tracked_series.items():
        tasks.append(
            partial(
                _handle_series,
                session,
                series_url,
                series_details,
                epub_generation_options,
                is_sync,
                new_synced,
                is_whole_volume,
            )
        )
        series_details_a.append(series_details)

    results = await bag(tasks)

    num_updated = 0
    num_errors = 0
    is_tracking_updated = False
    update_result: UpdateResult
    for i, update_result in enumerate(results):
        if not update_result.is_considered:
            continue

        series_details = series_details_a[i]

        if update_result.is_updated:
            num_updated += 1

        if update_result.is_error:
            num_errors += 1
        else:
            # the update of tracking has some conditions besides
            # just the series udpated
            is_tracking_updated_for_series = _update_tracking_data(
                series_details, update_result.series, update_result
            )
            if is_tracking_updated_for_series:
                is_tracking_updated = True

    if num_errors > 0:
        console.error("Some series could not be updated!")

    emoji = ""
    if console.is_advanced():
        emoji = "\u2728 "

    if num_updated == 0 and num_errors == 0:
        # second clause => all in error
        console.info(
            f"{emoji}All series are already up to date!",
            style="success",
        )

    if num_updated > 0:
        console.info(
            f"{emoji}{num_updated} series sucessfully updated!",
            style="success",
        )

    return is_tracking_updated


def _update_tracking_data(series_details, series_meta, update_result):
    if not (update_result.is_updated or update_result.is_force_set_updated):
        return False

    parts = core.all_parts_meta(series_meta)
    assert bool(parts)

    last_part = parts[-1]
    pn = spec.to_relative_spec_from_part(last_part)
    pdate = last_part.raw_data.launch

    series_details.part_date = pdate
    series_details.part = pn

    return True


async def _handle_series(
    session,
    series_url,
    series_details,
    epub_generation_options,
    is_sync,
    new_synced,
    is_whole_volume,
):
    series_meta = None
    try:
        if is_sync and series_url not in new_synced:
            return UpdateResult(is_considered=False)

        jnc_resource = jncweb.resource_from_url(series_url)
        series_meta = await core.resolve_series(session, jnc_resource)

        # generate from the start if the series is newly sync
        is_force_from_beginning = is_sync
        update_result = await _create_epub_for_new_parts(
            session,
            series_details,
            series_meta,
            epub_generation_options,
            is_whole_volume,
            is_force_from_beginning,
        )

        if update_result.is_updated:
            emoji = ""
            if console.is_advanced():
                emoji = "\u2714 "
            console.info(
                f"{emoji}The series '[highlight]{series_meta.raw_data.title}[/]' has "
                "been updated!",
                style="success",
            )

        return update_result

    except (trio.MultiError, Exception) as ex:
        if series_meta and series_meta.raw_data:
            title = series_meta.raw_data.title
        else:
            title = series_url

        emoji = ""
        if console.is_advanced():
            emoji = "\u274C "
        # FIXME show the user some feedback as to the nature of the error
        console.error(
            f"{emoji}Error updating '{title}'! "
            "(run 'jncep -d update' for more details)",
        )
        logger.debug(f"Error _handle_series: {ex}", exc_info=sys.exc_info())
        # series_meta may be None if error during retrieval
        return UpdateResult(is_error=True)


# TODO too complex ; refactor
async def _create_epub_for_new_parts(
    session,
    series_details,
    series_meta,
    epub_generation_options,
    is_whole_volume=False,
    is_force_from_beginning=False,
):
    # sufficient for here
    # quick to check
    await track.fill_meta_last_part(session, series_meta)
    parts = core.all_parts_meta(series_meta)

    if series_details.part == 0 or is_force_from_beginning:
        # Firt clause: special processing : means there was no part available when the
        # series was started tracking

        # still no part ?
        if not parts:
            return UpdateResult(is_updated=False)
        else:
            console.info(
                f"The series '[highlight]{series_meta.raw_data.title}[/]' "
                "will be updated..."
            )

            # complete series from beginning
            await core.fill_meta(session, series_meta)

            part_filter = partial(core.is_part_available, session.now)

            (
                volumes_to_download,
                parts_to_download,
            ) = core.relevant_volumes_and_parts_for_content(series_meta, part_filter)
            volumes_for_cover = core.relevant_volumes_for_cover(
                volumes_to_download, epub_generation_options.is_by_volume
            )

            await core.fill_covers_and_content(
                session, volumes_for_cover, parts_to_download
            )
            await core.create_epub(
                series_meta,
                volumes_to_download,
                parts_to_download,
                epub_generation_options,
            )

            return UpdateResult(series_meta, is_updated=True)
    else:

        if not series_details.part_date:
            # TODO test this branch
            # if here => old format, first lookup date of last part and use that
            # maybe still useful for stalled series so keep it
            part_spec = spec.analyze_part_specs(series_details.part)
            await core.fill_meta(session, series_meta, part_spec.has_volume)
            parts = core.all_parts_meta(series_meta)
            for part in parts:
                if part_spec.has_part(part):
                    # will be filled if the part still exists (it should)
                    # TODO case it doesn't ? eg tracked.json filled by hand
                    last_update_part = part
                    break
            # in UTC
            last_update_date = last_update_part.raw_data.launch
        else:
            # new format : date is recorded
            last_update_date = series_details.part_date

        last_update_date = dateutil.parser.parse(last_update_date)

        # TOC is below a part in the Labs API
        last_part = parts[-1]
        toc = await session.api.fetch_data("parts", last_part.part_id, "toc")
        # weird struct for the response : toc.parts has pagination struct (but all
        # parts seem to be there anyway) and parts property in turn
        # parts_to_download is a lsit of Parts but has reduced data (no num, no series
        # no volume)
        parts_release_after_date = _filter_parts_released_after_date(
            last_update_date, toc.parts.parts
        )

        if not parts_release_after_date:
            # not updated
            return UpdateResult(series_meta, is_updated=False)

        # filling in the series ; necessary to check availability
        for part in parts_release_after_date:
            part.series = series_meta

        available_parts_to_download = [
            part
            for part in parts_release_after_date
            if core.is_part_available(session.now, part)
        ]

        if not available_parts_to_download:
            console.warning(
                f"All updated parts for '[highlight]{series_meta.raw_data.title}[/]' "
                "have expired!"
            )
            # not updated but the series will still have its tracking data changed
            # in tracking config ; if not, the message above will always be displayed
            # should be rare (if updating often) ; also first part is preview
            # so even rarer
            return UpdateResult(
                series=series_meta, is_updated=False, is_force_set_updated=True
            )

        console.info(
            f"The series '[highlight]{series_meta.raw_data.title}[/]' will "
            "be updated..."
        )

        # after the to update notification
        if len(available_parts_to_download) != len(parts_release_after_date):
            console.warning(
                f"Some parts for '[highlight]{series_meta.raw_data.title}[/]' have "
                "expired!"
            )

        # we need the volumes for titles and covers
        # so fetch the meta to prepare
        await fill_meta_for_update(session, series_meta, available_parts_to_download)
        # normally if in this branch parts should not be empty
        parts = core.all_parts_meta(series_meta)

        parts_id_to_download = set(
            (part.part_id for part in available_parts_to_download)
        )

        # availability alread tested
        def simple_part_filter(part):
            return part.part_id in parts_id_to_download

        (
            volumes_to_download,
            parts_to_download,
        ) = core.relevant_volumes_and_parts_for_content(series_meta, simple_part_filter)

        if is_whole_volume:
            # second pass : filter on the volumes_to_download
            # all the parts of those volumes must be downloaded
            volumes_id_to_download = set((v.volume_id for v in volumes_to_download))

            def whole_volume_part_filter(part):
                return (
                    part.volume.volume_id in volumes_id_to_download
                    and core.is_part_available(session.now, part)
                )

            (
                volumes_to_download,
                parts_to_download,
            ) = core.relevant_volumes_and_parts_for_content(
                series_meta, whole_volume_part_filter
            )

        volumes_for_cover = core.relevant_volumes_for_cover(
            volumes_to_download, epub_generation_options.is_by_volume
        )

        tasks = [
            partial(
                core.fill_covers_and_content,
                session,
                volumes_for_cover,
                parts_to_download,
            ),
            partial(
                core.fill_num_parts_for_volumes,
                session,
                series_meta,
                volumes_to_download,
            ),
        ]
        await bag(tasks)

        await core.create_epub(
            series_meta,
            volumes_to_download,
            parts_to_download,
            epub_generation_options,
        )

        return UpdateResult(series_meta, is_updated=True)


async def fill_meta_for_update(session, series, parts_to_download):
    await core.fill_volumes_meta(session, series)
    volumes = series.volumes

    parts_id_to_download = set((part.part_id for part in parts_to_download))

    # in order not to do too many requests:
    # first try last 2 volumes => in most update scenarios (if updated frequently),
    # this should be enough
    # then do the rest if not sufficient
    last_2_volumes = volumes[-2:]
    rest_volumes = volumes[:-2]
    await core.fill_parts_meta_for_volumes(session, last_2_volumes)
    for volume in last_2_volumes:
        for part in volume.parts:
            # check if has part before the reference date
            # => means the parts relased after are all there
            # (except possibly when multiple volumes > 2 published at the same time)
            if part.part_id in parts_id_to_download:
                parts_id_to_download.remove(part.part_id)
        if len(parts_id_to_download) == 0:
            # found them all
            break
    else:
        # some parts to download have not been found in the last 2 volumes
        # just give up and request everything
        await core.fill_parts_meta_for_volumes(session, rest_volumes)


def _is_released_after_date(date, part_date_s):
    launch_date = dateutil.parser.parse(part_date_s)
    return launch_date > date


def _filter_parts_released_after_date(date, parts_raw):
    # parts is the raw Part struct from JNC Labs API
    parts_to_download = []
    for part_raw in parts_raw:
        # all date strings are in ISO format
        # so no need to parse really
        # parsing just to be safe
        # in case different shape like ms part or not (which throws str comp off)
        if _is_released_after_date(date, part_raw.launch):
            # num in volume is unknown at this point
            parts_to_download.append(model.Part(part_raw, part_raw.legacyId, -1))

    return parts_to_download



from functools import wraps
import json
import logging

from addict import Dict as Addict
import asks
import trio

from . import jncweb, utils
from .utils import deep_freeze

logger = logging.getLogger(__name__)
console = utils.getConsole()

CDN_IMG_URL_BASE = "https://d2dq7ifhe7bu0f.cloudfront.net"
LABS_API_JNC_URL_BASE = "https://labs.j-novel.club"
JNC_API_URL_BASE = "https://api.j-novel.club"

LABS_API_JNC_PATH_BASE = "/app/v1"
JNC_API_PATH_BASE = "/api"

COMMON_API_HEADERS = {"accept": "application/json", "content-type": "application/json"}
COMMON_LABS_API_PARAMS = {"format": "json"}


class InvalidCDNRequestException(Exception):
    pass


def with_cache(f):
    @wraps(f)
    async def wrapper(*args, **kwargs):
        # with_cache used only for JNCLabsAPI so fine
        api = args[0]
        if not hasattr(api, "__cache"):
            # cache scoped to api instance
            api.__cache = ({}, {})
        cache, events = api.__cache

        key = (*args, *kwargs.items())
        while True:
            if key in events:
                # query running
                # wait for it to finish
                logger.debug(f"{key} in events")
                event = events[key]
                await event.wait()
                if key in cache:
                    logger.debug(f"Cache hit {key}")
                    return _copy_or_raw(cache[key])
                # must have been error in the query
                # retry from the beginning in case
                # multiple are waiting
                # TODO raise instead ?
                continue

            event = trio.Event()
            events[key] = event

            try:
                response = await f(*args, **kwargs)
                cache[key] = _copy_or_raw(response)
                return response
            except Exception:
                del events[key]
                raise
            finally:
                # wake up the tasks waiting
                event.set()

    return wrapper


def _copy_or_raw(data):
    if type(data) is Addict:
        cp = data.deepcopy()
        # alway refreeze : fine in this context
        deep_freeze(cp)
        return cp
    # an image content ; won't be modifed so can be shared
    return data


class JNCLabsAPI:
    def __init__(
        self,
        jnc_connections=10,
        cdn_connections=20,
        jncweb_connections=10,
        labs_api_default_timeout=10,
        jnc_api_default_timeout=10,
        cdn_default_timeout=20,
        jncweb_default_timeout=10,
        connection_timeout=10,
    ):
        # jnc_connections params used for both labs and api (but only 1 req at a time
        # to api)
        self.jnc_api_session = asks.Session(
            JNC_API_URL_BASE, connections=jnc_connections, headers=COMMON_API_HEADERS
        )
        self.labs_api_session = asks.Session(
            LABS_API_JNC_URL_BASE,
            connections=jnc_connections,
            headers=COMMON_API_HEADERS,
        )
        # CDN_IMG_URL_BASE not really necessary
        self.cdn_session = asks.Session(CDN_IMG_URL_BASE, connections=cdn_connections)

        self.jncweb_session = asks.Session(
            jncweb.JNC_URL_BASE, connections=jncweb_connections
        )

        self.labs_api_default_timeout = labs_api_default_timeout
        self.jnc_api_default_timeout = jnc_api_default_timeout
        self.cdn_default_timeout = cdn_default_timeout
        self.jncweb_default_timeout = jncweb_default_timeout

        self.connection_timeout = connection_timeout

        self.token = None

    @property
    def is_logged_in(self):
        return self.token is not None

    def api_authentication(self):
        return {"authorization": self.token}

    def labs_authentication(self):
        return {"Authorization": f"Bearer {self.token}"}

    async def login(self, email, password):
        path = f"{LABS_API_JNC_PATH_BASE}/auth/login"
        payload = {"login": email, "password": password, "slim": True}
        params = {**COMMON_LABS_API_PARAMS}

        r = await self.labs_api_session.post(
            path=path,
            data=json.dumps(payload),
            params=params,
            connection_timeout=self.connection_timeout,
            timeout=self.labs_api_default_timeout,
        )
        r.raise_for_status()

        data = r.json()
        self.token = data["id"]

    async def logout(self):
        path = f"{LABS_API_JNC_PATH_BASE}/auth/logout"
        await self._call_labs_authenticated("POST", path)
        self.token = None

    @with_cache
    async def fetch_data(self, resource_type, slug_id, sub_resource="", skip=None):
        if sub_resource:
            sub_resource = f"/{sub_resource}"

        path = f"{LABS_API_JNC_PATH_BASE}/{resource_type}/{slug_id}{sub_resource}"
        params = {**COMMON_LABS_API_PARAMS}
        if skip is not None:
            params.update(skip=skip)

        logger.debug(f"LABS {path} skip={skip}")

        r = await self._call_labs_authenticated("GET", path, params=params)

        d = Addict(r.json())
        deep_freeze(d)
        return d

    async def paginate(self, func):
        skip = 0
        while True:
            j = await func(skip=skip)

            pagination = Addict(j.pop("pagination"))

            # besides pagination there is one other kv with the data we want
            items = list(j.values())
            for item in items[0]:
                yield item

            if pagination.lastPage:
                break
            skip += pagination.limit

    @with_cache
    async def fetch_content(self, slug_id, content_type):
        # not LABS_API base for embed queries
        path = f"/embed/{slug_id}/{content_type}"

        logger.debug(f"LABS EMBED {path}")

        r = await self._call_labs_authenticated("GET", path)
        return r.text

    async def _call_labs_authenticated(
        self, method, path, headers=None, params=None, **kwargs
    ):
        # TODO does almost nothing => remove ?
        auth = self.labs_authentication()
        r = await self._call_authenticated(
            self.labs_api_session,
            method,
            path,
            auth,
            headers,
            params,
            connection_timeout=self.connection_timeout,
            timeout=self.labs_api_default_timeout,
            **kwargs,
        )

        return r

    async def _call_api_authenticated(
        self, method, path, headers=None, params=None, **kwargs
    ):
        auth = self.api_authentication()
        if not headers:
            headers = COMMON_API_HEADERS
        else:
            headers = {**COMMON_API_HEADERS, **headers}

        path = f"{JNC_API_PATH_BASE}{path}"

        r = await self._call_authenticated(
            self.jnc_api_session,
            method,
            path,
            auth,
            headers,
            params,
            connection_timeout=self.connection_timeout,
            timeout=self.jnc_api_default_timeout,
            **kwargs,
        )

        return r

    async def _call_authenticated(
        self, session, method, path, auth, headers=None, params=None, **kwargs
    ):
        if not headers:
            headers = auth
        else:
            headers = {**auth, **headers}

        r = await session.request(
            method, path=path, headers=headers, params=params, **kwargs
        )
        r.raise_for_status()

        return r

    async def fetch_follows(self):
        path = "/users/me"
        qfilter = {"include": [{"serieFollows": "serie"}]}
        payload = {"filter": json.dumps(qfilter)}

        r = await self._call_api_authenticated("GET", path, params=payload)
        r.raise_for_status()

        me_data = Addict(r.json())
        followed_series = []
        for s in me_data.serieFollows:
            series = s.serie
            slug = series.titleslug
            # the metadata is not as complete as the usual (with fetch_metadata)
            # but it can still be useful to avoid a call later to the API
            jnc_resource = jncweb.JNCResource(
                jncweb.url_from_series_slug(slug),
                slug,
                True,
                jncweb.RESOURCE_TYPE_SERIES,
                series,
            )
            followed_series.append(jnc_resource)

        return followed_series

    async def follow_series(self, series_id):
        await self._set_follow(series_id, True)

    async def unfollow_series(self, series_id):
        await self._set_follow(series_id, False)

    async def _set_follow(self, series_id, is_follow):
        action = "follow" if is_follow else "unfollow"
        path = f"/users/me/{action}"

        payload = {"serieId": series_id, "serieType": 1}

        r = await self._call_api_authenticated("POST", path, json=payload)
        r.raise_for_status()

    @with_cache
    async def fetch_url(self, url: str):
        if not url.startswith(CDN_IMG_URL_BASE):
            raise InvalidCDNRequestException(
                f"{url} doesn't start with {CDN_IMG_URL_BASE}"
            )

        # for CDN images
        logger.debug(f"IMAGE {url}")
        r = await self.cdn_session.get(
            url=url,
            connection_timeout=self.connection_timeout,
            timeout=self.cdn_default_timeout,
        )
        r.raise_for_status()
        # should be JPEG
        # TODO check ?
        return r.content

    @with_cache
    async def fetch_jnc_webpage(self, series_slug):
        url = jncweb.url_from_series_slug(series_slug)
        logger.debug(f"JNCWEB {url}")
        r = await self.jncweb_session.get(
            url=url,
            connection_timeout=self.connection_timeout,
            timeout=self.jncweb_default_timeout,
        )
        r.raise_for_status()
        return r.text



import logging
import re
from urllib.parse import urlparse, urlunparse

import attr

logger = logging.getLogger(__name__)

JNC_URL_BASE = "https://j-novel.club"

RESOURCE_TYPE_SERIES = "SERIES"
RESOURCE_TYPE_VOLUME = "VOLUME"
RESOURCE_TYPE_PART = "PART"


class BadWebURLError(Exception):
    pass


@attr.s
class JNCResource:
    url = attr.ib()
    slug = attr.ib()
    is_new_website = attr.ib()
    resource_type = attr.ib()
    # used when fetching the follows and building a JNCRes
    follow_raw_data = attr.ib(None)

    def __str__(self):
        pu = urlparse(self.url)
        # remove scheme + domain
        return urlunparse(("", "", *pu[2:]))


def resource_from_url(url):
    pu = urlparse(url)

    if pu.scheme == "":
        raise BadWebURLError(f"Not a URL: {url}")

    # try legacy URL first
    # path is: /c/<slug>/...  or /v/<slug>/... or /s/<slug>/...
    # for c_hapter, v_olume, s_erie
    m = re.match(r"^/(v|c|s)/(.+?)(?:(?=/)|$)", pu.path)
    if m:
        return JNCResource(url, m.group(2), False, _to_const_legacy(m.group(1)))
    else:
        # new site
        # new site changed titles to series in URL
        # so process both
        s_re = r"^/(?:series|titles)/(.+?)(?:(?=/)|$)"
        c_re = r"^/read/(.+?)(?:(?=/)|$)"
        v_re = r"^volume-(\d+)$"

        m = re.match(s_re, pu.path)
        if m:
            series_slug = m.group(1)
            if not pu.fragment:
                return JNCResource(url, series_slug, True, RESOURCE_TYPE_SERIES)
            m = re.match(v_re, pu.fragment)
            if m:
                # tuple with volume
                return JNCResource(
                    url, (series_slug, int(m.group(1))), True, RESOURCE_TYPE_VOLUME
                )
        else:
            m = re.match(c_re, pu.path)
            if m:
                return JNCResource(url, m.group(1), True, RESOURCE_TYPE_PART)

    raise BadWebURLError(f"Invalid path for URL: {url}")


def url_from_series_slug(series_slug):
    # new URL
    return f"{JNC_URL_BASE}/series/{series_slug}"


def to_new_website_series_url(series_url):
    # supports legacy URLs + new
    jnc_resource = resource_from_url(series_url)
    # outputs new URL
    new_series_url = url_from_series_slug(jnc_resource.slug)
    return new_series_url


def _to_const_legacy(req_type):
    if req_type == "c":
        return RESOURCE_TYPE_PART
    elif req_type == "v":
        return RESOURCE_TYPE_VOLUME
    else:
        return RESOURCE_TYPE_SERIES



from functools import partial, wraps
import logging
from typing import List

import attr
import outcome
import trio

logger = logging.getLogger(__name__)


def handle_BaseExceptions(exc):
    if (
        isinstance(exc, SystemExit)
        or isinstance(exc, KeyboardInterrupt)
        or isinstance(exc, GeneratorExit)
    ):
        return exc
    if isinstance(exc, trio.MultiError):
        for ex in exc.exceptions:
            base_ex = handle_BaseExceptions(ex)
            if base_ex:
                return base_ex
    return None


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return trio.run(partial(f, *args, **kwargs))
        except trio.MultiError as ex:
            # TODO does this make sense ???
            base_ex = handle_BaseExceptions(ex)
            if base_ex:
                raise base_ex
            # just the first
            raise ex.exceptions[0]

    return wrapper


# taken from trio-future (but needs py > 3.7.0) so done here
@attr.s
class Future:
    result_chan: trio.abc.ReceiveChannel = attr.ib()

    async def get(self):
        future_outcome = await self.outcome()
        return future_outcome.unwrap()

    async def outcome(self) -> outcome.Outcome:
        try:
            async with self.result_chan:
                return await self.result_chan.receive()
        except trio.ClosedResourceError:
            raise RuntimeError(
                "Trio resource closed (did you try to call outcome twice on this "
                "future?"
            )


async def bag(async_fns):
    # background and gather
    async with trio.open_nursery() as n:
        f_tasks = []
        for async_fn in async_fns:
            f_task = background(n, async_fn)
            f_tasks.append(f_task)

        results = await gather(n, f_tasks).get()
        return results


def background(nursery: trio.Nursery, async_fn) -> Future:
    send_chan, recv_chan = trio.open_memory_channel(1)

    async def producer():
        return_val = await outcome.acapture(async_fn)
        # Shield sending the result from parent cancellation.
        with trio.CancelScope(shield=True):
            async with send_chan:
                await send_chan.send(return_val)

    nursery.start_soon(producer)
    return Future(recv_chan)


def gather(nursery: trio.Nursery, futures: List[Future]) -> Future:
    result_list = [None] * len(futures)
    parent_send_chan, parent_recv_chan = trio.open_memory_channel(0)
    child_send_chan, child_recv_chan = trio.open_memory_channel(0)

    async def producer():
        async with child_send_chan:
            for i in range(len(futures)):
                nursery.start_soon(child_producer, i, child_send_chan.clone())

    async def child_producer(i: int, out_chan):
        async with futures[i].result_chan:
            return_val = await futures[i].result_chan.receive()
            result_list[i] = return_val
            async with out_chan:
                await out_chan.send(i)

    async def receiver():
        async with child_recv_chan:
            async for i in child_recv_chan:
                # Just consume all results from the channel until exhausted
                pass
        # And then wrap up the result and push it to the parent channel
        errors = [e.error for e in result_list if isinstance(e, outcome.Error)]
        if len(errors) > 0:
            result = outcome.Error(trio.MultiError(errors))
        else:
            result = outcome.Value([o.unwrap() for o in result_list])
        async with parent_send_chan:
            await parent_send_chan.send(result)

    # Start parent producer, which will in turn start all children
    # (doing this inside the nursery because it needs to act async)
    nursery.start_soon(producer)
    nursery.start_soon(receiver)
    return Future(parent_recv_chan)



import inspect
import logging
import re
import sys
import unicodedata

from addict import Dict as Addict
import rich.console
import rich.theme

logger = logging.getLogger(__name__)


def setup_logging(is_debug, package=__package__):
    if not logging.getLogger().handlers:
        # not needed if coloredlogs is used
        format = "%(asctime)s %(name)-12s %(levelname)-8s %(message)s"
        logging.basicConfig(stream=sys.stdout, level=logging.INFO, format=format)

    # coloredlogs changes the level of the handler
    logging.getLogger().handlers[0].setLevel(logging.NOTSET)

    logger = logging.getLogger(package)
    if is_debug:
        logger.setLevel(logging.DEBUG)
        # keep debug console (ie logging)
        # or issues mixing the print and the logs with Rich
    else:
        logger.setLevel(logging.INFO)
        getConsole().console = RichConsole()


def tryint(val):
    try:
        return int(val)
    except Exception:
        return None


def to_yn(b):
    return "yes" if b else "no"


def to_safe_filename(name):
    name = "".join(
        c for c in unicodedata.normalize("NFD", name) if unicodedata.category(c) != "Mn"
    )
    safe = re.sub(r"[^0-9a-zA-Z_]+", "_", name)
    safe = safe.strip("_")
    return safe


def module_info():
    # for main module : its __name__ is __main__
    # so find out its real name
    frm = inspect.stack()[1]
    mod = inspect.getmodule(frm[0])
    return mod.__spec__.name


def deep_freeze(data):
    if type(data) is Addict:
        data.freeze()
        for value in data.values():
            if type(value) is list:
                for v in value:
                    deep_freeze(v)


rich_theme = rich.theme.Theme(
    {
        "success": "green",
        "warning": "yellow",
        "error": "red",
        "highlight": "magenta",
    }
)


class RichConsole:
    def __init__(self):
        self.console = rich.console.Console(
            highlight=False, theme=rich_theme, soft_wrap=True
        )
        self._status = None

    def info(self, *args, **kwargs):
        self.console.print(*args, **kwargs)

    def warning(self, *args, **kwargs):
        self.console.print(*args, **{"style": "warning", **kwargs})

    def error(self, *args, **kwargs):
        self.console.print(*args, **{"style": "error", **kwargs})

    def status(self, message, **kwargs_spinner_style):
        if not self._status:
            default = {}
            if self.is_advanced():
                default = {"spinner": "dots"}
            else:
                default = {"spinner": "line", "refresh_per_second": 6}
            st_args = {**default, **kwargs_spinner_style}

            self._status = self.console.status(message, **st_args)
            self._status.start()
        else:
            self._status.update(message, **kwargs_spinner_style)

    def stop_status(self):
        if self._status:
            self._status.stop()

    def log(self, *args, **kwargs):
        self.console.log(*args, **kwargs)

    def is_advanced(self):
        return not self.console.legacy_windows


class DebugConsole:
    def __init__(self):
        pass

    def info(self, message, *args, **kwargs):
        logger.info(message)

    def warning(self, message, *args, **kwargs):
        logger.warning(message)

    def error(self, message, *args, **kwargs):
        logger.error(message)

    def status(self, message, **kwargs_spinner_style):
        logger.warning(message)

    def stop_status(self):
        pass

    def log(self, message, *args, **kwargs):
        logger.info(message)

    def is_advanced(self):
        return False


class RootConsole:
    def __init__(self):
        # default
        self.console = DebugConsole()

    def info(self, message, *args, **kwargs):
        self.console.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        self.console.warning(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        self.console.error(message, *args, **kwargs)

    def status(self, message, **kwargs_spinner_style):
        self.console.status(message, **kwargs_spinner_style)

    def stop_status(self):
        self.console.stop_status()

    def log(self, message, *args, **kwargs):
        self.console.log(message, *args, **kwargs)

    def is_advanced(self):
        return self.console.is_advanced()


ROOT_CONSOLE = RootConsole()


def getConsole():
    return ROOT_CONSOLE



import attr


@attr.s
class Series:
    raw_data = attr.ib()
    series_id = attr.ib()
    volumes = attr.ib(None)


@attr.s
class Volume:
    raw_data = attr.ib()
    volume_id = attr.ib()
    num = attr.ib()
    parts = attr.ib(None)
    is_analyzed = attr.ib(False)
    cover = attr.ib(None)
    series = attr.ib(None)
    parts_count = attr.ib(None)


@attr.s
class Part:
    raw_data = attr.ib()
    part_id = attr.ib()
    num_in_volume = attr.ib()
    volume = attr.ib(None)
    series = attr.ib(None)
    content = attr.ib(None)
    images = attr.ib(None)

    epub_content = attr.ib(None)


@attr.s
class Image:
    url = attr.ib()
    content = attr.ib(None)
    part = attr.ib(None)
    local_filename = attr.ib(None)

    order_in_part = attr.ib(None)
